/*	$OpenBSD: hil.c,v 1.7 2011/08/18 19:37:26 miod Exp $	*/
/*	$NetBSD: hil.c,v 1.2 1997/04/14 19:00:10 thorpej Exp $	*/

/*
 * Copyright (c) 1997 Jason R. Thorpe.  All rights reserved.
 * Copyright (c) 1988 University of Utah.
 * Copyright (c) 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * the Systems Programming Group of the University of Utah Computer
 * Science Department.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * from: Utah Hdr: hil.c 1.1 89/08/22
 *
 *	@(#)hil.c	8.1 (Berkeley) 6/10/93
 */

/*
 * HIL keyboard routines for the standalone ITE.
 */

#if defined(ITECONSOLE) && defined(HIL_KEYBOARD)

#include <lib/libsa/stand.h>

#include "samachdep.h"
#include "hilreg.h"
#include "itevar.h"
#include "kbdvar.h"

#ifndef SMALL

#define	ESC	'\033'
#define	DEL	'\177'

/*
 * HIL cooked keyboard keymaps.
 * Supports only unshifted, shifted and control keys.
 */
char	hilkbd_keymap[] = {
	0,	'`',	'\\',	ESC,	0,	DEL,	0,	0,
	'\n',	'\t',	0,	0,	0,	0,	0,	0,
	0,	'\n',	0,	0,	0,	0,	0,	0,
	0,	'\t',	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	'\b',	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	ESC,	'\r',	0,	'\n',	'0',	'.',	',',	'+',
	'1',	'2',	'3',	'-',	'4',	'5',	'6',	'*',
	'7',	'8',	'9',	'/',	'E',	'(',	')',	'^',
	'1',	'2',	'3',	'4',	'5',	'6',	'7',	'8',
	'9',	'0',	'-',	'=',	'[',	']',	';',	'\'',
	',',	'.',	'/',	'\040',	'o',	'p',	'k',	'l',
	'q',	'w',	'e',	'r',	't',	'y',	'u',	'i',
	'a',	's',	'd',	'f',	'g',	'h',	'j',	'm',
	'z',	'x',	'c',	'v',	'b',	'n',	0,	0
};

char	hilkbd_shiftmap[] = {
	0,	'~',	'|',	DEL,	0,	DEL,	0,	0,
	'\n',	'\t',	0,	0,	0,	0,	0,	0,
	0,	'\n',	0,	0,	0,	0,	0,	0,
	0,	'\t',	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	DEL,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	ESC,	'\r',	0,	'\n',	'0',	'.',	',',	'+',
	'1',	'2',	'3',	'-',	'4',	'5',	'6',	'*',
	'7',	'8',	'9',	'/',	'`',	'|',	'\\',	'~',
	'!',	'@',	'#',	'$',	'%',	'^',	'&',	'*',
	'(',	')',	'_',	'+',	'{',	'}',	':',	'\"',
	'<',	'>',	'?',	'\040',	'O',	'P',	'K',	'L',
	'Q',	'W',	'E',	'R',	'T',	'Y',	'U',	'I',
	'A',	'S',	'D',	'F',	'G',	'H',	'J',	'M',
	'Z',	'X',	'C',	'V',	'B',	'N',	0,	0
};

char	hilkbd_ctrlmap[] = {
	0,	'`',	'\034',	ESC,	0,	DEL,	0,	0,
	'\n',	'\t',	0,	0,	0,	0,	0,	0,
	0,	'\n',	0,	0,	0,	0,	0,	0,
	0,	'\t',	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	'\b',	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	ESC,	'\r',	0,	'\n',	'0',	'.',	',',	'+',
	'1',	'2',	'3',	'-',	'4',	'5',	'6',	'*',
	'7',	'8',	'9',	'/',	'E',	'(',	')',	'\036',
	'1',	'2',	'3',	'4',	'5',	'6',	'7',	'8',
	'9',	'0',	'-',	'=',	'\033',	'\035',	';',	'\'',
	',',	'.',	'/',	'\040',	'\017',	'\020',	'\013',	'\014',
	'\021',	'\027',	'\005',	'\022',	'\024',	'\031',	'\025',	'\011',
	'\001',	'\023',	'\004',	'\006',	'\007',	'\010',	'\012',	'\015',
	'\032',	'\030',	'\003',	'\026',	'\002',	'\016',	0,	0
};

int
hilkbd_getc()
{
	int status, c;
	struct hil_dev *hiladdr = HILADDR;

	status = hiladdr->hil_stat;
	if ((status & HIL_DATA_RDY) == 0)
		return(0);
	c = hiladdr->hil_data;
	switch ((status >> HIL_SSHIFT) & HIL_SMASK) {
	case KBD_SHIFT:
		c = hilkbd_shiftmap[c & KBD_CHARMASK];
		break;
	case KBD_CTRL:
		c = hilkbd_ctrlmap[c & KBD_CHARMASK];
		break;
	case KBD_KEY:
		c = hilkbd_keymap[c & KBD_CHARMASK];
		break;
	default:
		c = 0;
		break;
	}
	return(c);
}
#endif /* SMALL */

void
hilkbd_nmi()
{
	struct hil_dev *hiladdr = HILADDR;

	HILWAIT(hiladdr);
	hiladdr->hil_cmd = HIL_CNMT;
	HILWAIT(hiladdr);
	hiladdr->hil_cmd = HIL_CNMT;
	HILWAIT(hiladdr);
	printf("\nboot interrupted\n");
}

int
hilkbd_init()
{
	struct hil_dev *hiladdr = HILADDR;

	/*
	 * Determine the existence of a HIL keyboard.
	 */
	HILWAIT(hiladdr);
	hiladdr->hil_cmd = HIL_READKBDSADR;
	HILDATAWAIT(hiladdr);
	if (hiladdr->hil_data == 0)
		return (0);

	HILWAIT(hiladdr);
	hiladdr->hil_cmd = HIL_SETARR;
	HILWAIT(hiladdr);
	hiladdr->hil_data = ar_format(KBD_ARR);
	HILWAIT(hiladdr);
	hiladdr->hil_cmd = HIL_INTON;

	return (1);
}
#endif /* ITECONSOLE && HIL_KEYBOARD */
